import path from 'node:path';
import micromatch from 'micromatch';
import { ROOT_WORKSPACE_NAME } from './constants.js';
import * as plugins from './plugins/index.js';
import { LocalConfiguration } from './types/validate.js';
import { arrayify } from './util/array.js';
import { ConfigurationError } from './util/errors.js';
import { findFile, loadJSON } from './util/fs.js';
import { _dirGlob } from './util/glob.js';
import parsedArgs from './util/parseArgs.js';
import { resolveIncludedIssueTypes } from './util/resolveIncludedIssueTypes.js';
import { workspaceSorter } from './util/sort.js';
const { values: { config: rawConfigArg, workspace: rawWorkspaceArg, include = [], exclude = [], strict: isStrict = false, production: isProduction = false, }, } = parsedArgs;
const defaultConfig = {
    include: [],
    exclude: [],
    ignoreBinaries: [],
    ignoreFiles: [],
    ignoreWorkspaces: [],
    workspaces: {
        [ROOT_WORKSPACE_NAME]: {
            entryFiles: ['index.{js,ts,tsx}', 'src/index.{js,ts,tsx}'],
            projectFiles: ['**/*.{js,ts,tsx}'],
            ignore: [],
            paths: {},
        },
    },
};
const PLUGIN_NAMES = Object.keys(plugins);
export default class ConfigurationChief {
    cwd = process.cwd();
    config;
    manifestPath;
    manifest;
    constructor({ cwd }) {
        this.cwd = cwd ?? this.cwd;
        this.config = defaultConfig;
    }
    async loadLocalConfig() {
        const manifestPath = await findFile(this.cwd, 'package.json');
        const manifest = manifestPath && (await loadJSON(manifestPath));
        if (!manifestPath || !manifest) {
            throw new ConfigurationError('Unable to find package.json');
        }
        this.manifestPath = manifestPath;
        this.manifest = manifest;
        const configFilePath = rawConfigArg ?? 'knip.json';
        const resolvedConfigFilePath = (await findFile(this.cwd, configFilePath)) ?? (!rawConfigArg && (await findFile(this.cwd, configFilePath + 'c')));
        if (rawConfigArg && !resolvedConfigFilePath && !manifest.knip) {
            throw new ConfigurationError(`Unable to find ${rawConfigArg} or package.json#knip`);
        }
        const rawLocalConfig = resolvedConfigFilePath ? await loadJSON(resolvedConfigFilePath) : manifest.knip;
        if (rawLocalConfig) {
            this.config = this.normalize(LocalConfiguration.parse(rawLocalConfig));
        }
    }
    normalize(rawLocalConfig) {
        const workspaces = rawLocalConfig.workspaces ?? {
            [ROOT_WORKSPACE_NAME]: {
                ...rawLocalConfig,
            },
        };
        const include = rawLocalConfig.include ?? defaultConfig.include;
        const exclude = rawLocalConfig.exclude ?? defaultConfig.exclude;
        const ignoreBinaries = rawLocalConfig.ignoreBinaries ?? defaultConfig.ignoreBinaries;
        const ignoreFiles = arrayify(rawLocalConfig.ignoreFiles ?? defaultConfig.ignoreFiles);
        const ignoreWorkspaces = rawLocalConfig.ignoreWorkspaces ?? defaultConfig.ignoreWorkspaces;
        return {
            include,
            exclude,
            ignoreBinaries,
            ignoreFiles,
            ignoreWorkspaces,
            workspaces: Object.entries(workspaces)
                .filter(([workspaceName]) => !ignoreWorkspaces.includes(workspaceName))
                .reduce((workspaces, workspace) => {
                const [workspaceName, workspaceConfig] = workspace;
                const entryFiles = arrayify(workspaceConfig.entryFiles);
                workspaces[workspaceName] = {
                    entryFiles,
                    projectFiles: arrayify(workspaceConfig.projectFiles ?? entryFiles),
                    ignore: arrayify(workspaceConfig.ignore),
                    paths: workspaceConfig.paths ?? {},
                };
                for (const [pluginName, pluginConfig] of Object.entries(workspaceConfig)) {
                    if (PLUGIN_NAMES.includes(pluginName)) {
                        const isObject = typeof pluginConfig !== 'string' && !Array.isArray(pluginConfig);
                        const config = isObject ? arrayify(pluginConfig.config) : arrayify(pluginConfig);
                        const entryFiles = isObject && 'entryFiles' in pluginConfig ? arrayify(pluginConfig.entryFiles) : [];
                        const projectFiles = isObject && 'projectFiles' in pluginConfig ? arrayify(pluginConfig.projectFiles) : entryFiles;
                        workspaces[workspaceName][pluginName] = {
                            config,
                            entryFiles,
                            projectFiles,
                        };
                    }
                }
                return workspaces;
            }, {}),
        };
    }
    async getManifestWorkspaces() {
        const { workspaces } = this.manifest ?? {};
        const patterns = workspaces ? (Array.isArray(workspaces) ? workspaces : workspaces.packages ?? []) : [];
        return _dirGlob({ patterns, cwd: this.cwd, ignore: this.config.ignoreWorkspaces });
    }
    getConfiguredWorkspaces() {
        return this.config.workspaces ? Object.keys(this.config.workspaces) : [];
    }
    async getActiveWorkspaces() {
        const manifestWorkspaces = await this.getManifestWorkspaces();
        const additionalWorkspaces = Object.keys(this.config.workspaces).filter(name => !name.includes('*') && !manifestWorkspaces.includes(name));
        const rootWorkspace = {
            name: ROOT_WORKSPACE_NAME,
            dir: this.cwd,
            config: this.getConfigForWorkspace(ROOT_WORKSPACE_NAME),
            ancestors: [],
        };
        if (manifestWorkspaces.length === 0 && !rawWorkspaceArg)
            return [rootWorkspace];
        if (rawWorkspaceArg) {
            return [
                rootWorkspace,
                {
                    name: rawWorkspaceArg,
                    dir: path.resolve(this.cwd, rawWorkspaceArg),
                    config: this.getConfigForWorkspace(rawWorkspaceArg),
                    ancestors: [ROOT_WORKSPACE_NAME],
                },
            ];
        }
        const workspaces = [...manifestWorkspaces, ...additionalWorkspaces];
        const activeWorkspaces = workspaces.filter(workspaceName => this.hasConfigForWorkspace(workspaceName));
        return activeWorkspaces.sort(workspaceSorter).map(name => ({
            name,
            dir: path.resolve(this.cwd, name),
            config: this.getConfigForWorkspace(name),
            ancestors: activeWorkspaces.reduce((ancestors, ancestorName) => {
                if (name === ancestorName)
                    return ancestors;
                if (ancestorName === ROOT_WORKSPACE_NAME || name.startsWith(ancestorName)) {
                    if (this.hasConfigForWorkspace(ancestorName)) {
                        ancestors.push(ancestorName);
                    }
                }
                return ancestors;
            }, []),
        }));
    }
    getConfigKeyForWorkspace(workspaceName) {
        const configuredWorkspaces = this.getConfiguredWorkspaces();
        return configuredWorkspaces
            .sort(workspaceSorter)
            .reverse()
            .find(pattern => micromatch.isMatch(workspaceName, pattern));
    }
    hasConfigForWorkspace(workspaceName) {
        return Boolean(this.getConfigKeyForWorkspace(workspaceName));
    }
    getConfigForWorkspace(workspaceName) {
        const key = this.getConfigKeyForWorkspace(workspaceName);
        if (key) {
            return this.config?.workspaces?.[key] ?? { entryFiles: [], projectFiles: [], paths: {}, ignore: [] };
        }
        return { entryFiles: [], projectFiles: [], paths: {}, ignore: [] };
    }
    resolveIncludedIssueTypes() {
        return resolveIncludedIssueTypes(include, exclude, {
            include: this.config.include ?? [],
            exclude: this.config.exclude ?? [],
            isProduction,
            isStrict,
        });
    }
}
