import path from 'node:path';
import { ts } from 'ts-morph';
import { debugLogFiles, debugLogSourceFiles } from './util/debug.js';
import { _createProject, partitionSourceFiles, _resolveSourceFileDependencies, _removeExternalSourceFiles, } from './util/project.js';
const compilerOptions = {
    allowJs: true,
    allowSyntheticDefaultImports: true,
    jsx: ts.JsxEmit.ReactJSX,
    esModuleInterop: true,
    skipDefaultLibCheck: true,
    skipLibCheck: true,
    target: ts.ScriptTarget.ES2015,
    module: ts.ModuleKind.CommonJS,
    moduleResolution: ts.ModuleResolutionKind.NodeJs,
};
const skipAddFiles = { skipAddingFilesFromTsConfig: true, skipFileDependencyResolution: true };
export default class ProjectPrincipal {
    projectOptions;
    tsConfigFilePath;
    entryPaths = new Set();
    projectPaths = new Set();
    constructor() {
        this.projectOptions = this.tsConfigFilePath
            ? { tsConfigFilePath: this.tsConfigFilePath, compilerOptions }
            : { compilerOptions };
    }
    addEntryPath(filePath) {
        this.entryPaths.add(filePath);
    }
    addProjectPath(filePath) {
        this.projectPaths.add(filePath);
    }
    addTypeScriptPaths(workspaceDir, paths) {
        for (const [key, entries] of Object.entries(paths)) {
            const workspacePaths = entries.map(entry => path.join(workspaceDir, entry));
            if (!this.projectOptions.compilerOptions)
                this.projectOptions.compilerOptions = {};
            if (!this.projectOptions.compilerOptions.paths)
                this.projectOptions.compilerOptions.paths = {};
            if (this.projectOptions.compilerOptions.paths[key]) {
                this.projectOptions.compilerOptions.paths[key].push(...workspacePaths);
            }
            else {
                this.projectOptions.compilerOptions.paths[key] = workspacePaths;
            }
        }
    }
    createProjects() {
        const production = _createProject({ ...this.projectOptions, ...skipAddFiles }, [...this.entryPaths]);
        const entryFiles = production.getSourceFiles();
        debugLogSourceFiles(1, `Included entry files`, entryFiles);
        _resolveSourceFileDependencies(production);
        const productionFiles = _removeExternalSourceFiles(production);
        debugLogSourceFiles(1, 'Included files resolved from entry files', productionFiles);
        const project = _createProject({ ...this.projectOptions, ...skipAddFiles }, [...this.projectPaths]);
        const projectFiles = project.getSourceFiles();
        debugLogSourceFiles(1, 'Included project files', projectFiles);
        return { entryFiles, productionFiles, projectFiles };
    }
    settleFiles() {
        const { entryFiles, productionFiles, projectFiles } = this.createProjects();
        const [usedProductionFiles, unreferencedProductionFiles] = partitionSourceFiles(projectFiles, productionFiles);
        const [usedEntryFiles, usedNonEntryFiles] = partitionSourceFiles(usedProductionFiles, entryFiles);
        debugLogSourceFiles(1, 'Used production files', usedProductionFiles);
        debugLogFiles(1, 'Unreferenced production files', unreferencedProductionFiles);
        debugLogSourceFiles(1, 'Used entry files', usedEntryFiles);
        debugLogFiles(1, 'Used non-entry files', usedNonEntryFiles);
        return { usedProductionFiles, unreferencedProductionFiles };
    }
}
